
-- ------------------------------------------------------------
-- Kuhn Poker versus a Nash optimal player
--
-- Observation space = 4 bits
-- Reward space      = 3 bits
-- Action space      = 1 bits
--
-- The agent is second to act.
-- If it plays optimally, it can expect to win 1/18 per hand.
-- ------------------------------------------------------------

-- program options
debug_mode = true;


-- initialise prng
seed = os.time()
math.randomseed(seed)
math.random(); math.random(); math.random();

-- player 1 strategy - setting gamma <- [0,1] gives different equilibrum strategies
gamma = 0.7
alpha = gamma / 3.0
beta  = (1.0 + gamma) / 3.0


-- global variables
king  =  3; queen =  2; jack  =  1; none  = -1
phase = 0
p1 = {}; p2 = {}

-- pretty printing of hidden state
pp = {}
pp[king]   = 'King'
pp[queen]  = 'Queen'
pp[jack]   = 'Jack'
pp['bet']  = 'bets'
pp['pass'] = 'passes'


-- encoding/decoding tables
encode_obs = {}
encode_obs['bet']  = '00'
encode_obs['pass'] = '11'
encode_obs[none]   = '00'
encode_obs[king]   = '01'
encode_obs[queen]  = '10'
encode_obs[jack]   = '11'

encode_rew = {}
encode_rew[0] = '000'
encode_rew[1] = '001'
encode_rew[3] = '011'
encode_rew[4] = '100'

decode_act = {}
decode_act['0'] = 'pass'
decode_act['1'] = 'bet'


-- reset the round
function reset_round ()

	p1.reward = -1
	p2.reward = -1

	p1.hole_card = math.random(3)
	p2.hole_card = p1.hole_card

	while (p2.hole_card == p1.hole_card) do
		p2.hole_card = math.random(3)
	end

	p1.action1 = p1_first_action()
	p1.action2 = 'none'

	dprint('p1=' .. pp[p1.hole_card] .. ', p2=' .. pp[p2.hole_card])
	dprint('p1 ' .. pp[p1.action1])
end


-- compute player 1's first action
function p1_first_action()

	local r = math.random()

	if p1.hole_card == jack then
		dprint('p1 deciding to bet, with probability: ' .. alpha)
		if (r < alpha) then return 'bet' else return 'pass' end
	elseif p1.hole_card == queen then
		return 'pass'
	else
		dprint('p1 deciding to bet, with probability: ' .. gamma)
		if (r < gamma) then return 'bet' else return 'pass' end
	end
end


-- compute player 1's second action
function p1_second_action()

	if p1.hole_card == jack then
		return 'pass'
	elseif p1.hole_card == queen then
		local r = math.random()
		dprint('p1 deciding to bet again, with probability: ' .. beta)
		if (r < beta) then return 'bet' else return 'pass' end
	else
		return 'bet'
	end
end


-- does player 1 need to make a second action?
function need_second_action()

	if (p1.action1 == p2.action) then
		return false
	elseif (p1.action1 == 'bet' and p2.action == 'pass') then
		return false
	else
		return true
	end

end


-- compute winner from a showdown
function showdown_winner()

	if (p1.hole_card > p2.hole_card) then
		return 1
	else
		return 2
	end
end


-- compute the winner of a hand
function hand_winner()

	local ante = 2
	local p1_input = 0; p1_input = 0

	-- handle all cases where a player wins by getting the opponent to fold
	if (p1.action1 == 'bet'  and p2.action == 'pass') then
		return 1
	end

	if (p1.action1 == 'pass' and p2.action == 'bet' and p1.action2 == 'pass') then
		return 2
	end

	-- otherwise a showdown happened
	return showdown_winner()

end


-- compute winnings for player 2
function p2_winnings()

	local invested = 1 -- player 2's ante

	-- compute total money in pot
	local pot = 2 -- add antes
	if (p1.action1 == 'bet') then pot = pot + 1 end
	if (p2.action  == 'bet') then
		pot = pot + 1
		invested = invested + 1
	end
	if (p1.action2 == 'bet') then pot = pot + 1 end

	if hand_winner() == 2 then
		return pot - invested + 2
	else
		return 2 - invested
	end
end


-- compute the reward for an agent based on the world state
function compute_reward ()

	if need_second_action() then
		p1.action2 = p1_second_action()
	end

	return p2_winnings()

end


-- compute the next observation based on the hidden state
function compute_observation ()

	local card_obs = encode_obs[p2.hole_card]
	local act_obs = encode_obs[p1.action1]

	return card_obs .. act_obs

end


-- debugging print
function dprint (str)

	if (debug_mode) then
		local f = io.output()
		io.output(stderr)
		print(str)
		io.output(f)
	end
end



-- main loop
line   = ''
rew = 0
hand_no = 1

dprint('p1 strategy: (' .. alpha .. ', ' .. beta .. ', ' .. gamma .. ')')

repeat

	dprint('')
	dprint('------------------------------')
	dprint('Hand ' .. hand_no)
	dprint('')

	reset_round()

	local obs = compute_observation()

	dprint('obs=' .. obs .. ', rew=' .. rew)

	local enc_rew = encode_rew[rew]

	print(obs .. enc_rew)
	io.flush()


	-- grab action
	line = io.read()
	if (line == '') then break end

	p2.action = decode_act[line];

	rew = compute_reward()

	dprint('p2 ' .. pp[p2.action])
	if p1.action2 ~= 'none' then
		dprint('p1 ' .. pp[p1.action2])
	end

	dprint('p' .. hand_winner() .. ' wins the hand.')
	dprint('p2 receives ' .. (rew-2) .. ' reward.')

	hand_no = hand_no + 1

until line == ''


