#include "utree.hpp"
#include "log.hpp"

#include <iostream>
#include <limits>
#include <fstream>

// boost includes
#include <boost/program_options.hpp>
#include <boost/random.hpp>
#include <boost/tokenizer.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>


#undef max

#include "log.hpp"

namespace po = boost::program_options;

/// malformed environment response exception
class BadPerceptException : public std::exception { 
	const char *what(void) const throw() { return "invalid percept"; }
};

// program details
static const char program_name[] = "A U-Tree Implementation";
static const char version[]      = "0.1";
static const char authors[]      = "Joel Veness";

// maximum age of Active-LZ agent
static const unsigned int MaxAgentAge = 1000000000;

// uniform [0..1) random number generation
static randsrc_t rng;
static randgen_t urng(rng);

// stores the application configuration options and their description
po::variables_map options;
static po::options_description options_desc("Usage", 100);

static std::auto_ptr<UTree> utree;


/* load program options into a variable map */
static void initOptions(int argc, char *argv[], po::variables_map &vm) {

    options_desc.add_options()

        ("help", "")

	    ("reward-bits",
		 po::value<unsigned int>()->default_value(1),
		 "how many bits are used to encode the reward signal")

	    ("observation-bits",
		 po::value<unsigned int>()->default_value(1),
         "how many bits are used to encode the observation signal")

		("agent-actions",
		 po::value<unsigned int>()->default_value(4),
		 "the number of distinct actions the agent can do")

		("learning-rate",
		 po::value<double>()->default_value(0.05),
		 "the learning rate for the value iteration step")

	    ("agent-load",
		 po::value<std::string>(),
		 "load a pre-existing agent from a file")

	    ("discount",
		 po::value<double>()->default_value(0.99),
		 "discount factor")

		("exploration",
		 po::value<double>()->default_value(0.0),
		 "probability of playing a random move")

	    ("fringe",
		 po::value<unsigned int>()->default_value(1),
		 "how far do we expand the fringe by?")

		("depth",
		 po::value<unsigned int>()->default_value(4),
		 "how far back in the past can a utile distinction be made?")

	    ("test-fringe",
		 po::value<unsigned int>()->default_value(100),
		 "how many cycles per fringe test")

		("explore-decay",
		 po::value<double>()->default_value(1.0),
		 "a value between [0.0,1.0] that defines the geometric decay of the exploration rate")

	    ("terminate-age",
		 po::value<age_t>()->default_value(MaxAgentAge),
		 "how many agent/environment cycles before the agent needs to close")

		("binary-io",
		 "use native (faster but non-portable) binary file i/o")
    ;

    po::store(po::parse_command_line(argc, argv, options_desc), vm);
    po::notify(vm);
}


/* display the program usage options and quit. */
static void showHelp(void) {

	std::cout << std::endl
	          << program_name << " " << version << ", by " << authors << std::endl
	          << "--------------------------------------------------------"
	          << std::endl << std::endl
	          << "The environment communicates percepts to the agent over stdin." << std::endl
	          << "The agent responds to the environment over stdout." << std::endl << std::endl
	          << "All communication channels are binary." << std::endl
	          << std::endl << options_desc << std::endl;
    exit(0);
}


/* checks for illegal combinations of configuration parameters,
   throwing an exception if an illegal combination is detected. */
static void processOptions(const po::variables_map &vm) {

	if (vm.count("help") > 0) showHelp();

	if (vm.count("observation-bits") > 0) {
		if (vm["observation-bits"].as<unsigned int>() < 1) {
			throw std::invalid_argument("the number of observation bits must be positive.");
		}
	}

	if (vm.count("reward-bits") > 0) {
		if (vm["reward-bits"].as<unsigned int>() < 1) {
			throw std::invalid_argument("the number of reward bits must be positive.");
		}
		if (vm["reward-bits"].as<unsigned int>() > 32) {
			throw std::invalid_argument("the number of reward bits must be 32 or less.");
		}
	}

	if (vm.count("learning-rate") > 0) {
		if (vm["learning-rate"].as<double>() <= 0.0 ) {
			throw std::invalid_argument("the learning rate must be positive");
		}
	}

	if (vm.count("discount") > 0) {
		double gamma = vm["discount"].as<double>();
		if (gamma <= 0.0 || gamma >= 1.0) {
			throw std::invalid_argument("the discount factor must be within (0..1)");
		}
	}

	if (vm.count("fringe") > 0) {
		unsigned int fringe = vm["fringe"].as<unsigned int>();
		if (fringe == 0) {
			throw std::invalid_argument("fringe must be positive");
		}
	}

	if (vm.count("test-fringe") > 0) {
		unsigned int fringe = vm["test-fringe"].as<unsigned int>();
		if (fringe == 0) {
			throw std::invalid_argument("test-fringe must be positive");
		}
	}
}


/* determine the number of bits needed to encode each action */
static unsigned int num_action_bits(void) {

	unsigned int actions_c = options["agent-actions"].as<unsigned int>();
	unsigned int actions_bits_c;

	for (unsigned int i = 1, c = 1; i < actions_c; i *= 2, c++) {
		actions_bits_c = c;
	}

	return actions_bits_c;
}


/* display some summary statistics for the current cycle */
static void logStatistics(const UTree &utree) {

	unsigned int n_actions = options["agent-actions"].as<unsigned int>();

	for (unsigned int a=0; a < n_actions; a++) {
		const action_model_t *am = utree.actionModel(a);
		if (am != NULL && am->count > 0) {
			double rew = am->reward / double(am->count);
			LINFO_ << "Action " << a << ": " << am->q_value << ", visits: " << am->count
				<< ", rew: " << rew;
		} else {
			LINFO_ << "Action " << a << ": ?";
		}
	}
}


/* save the state of an agent to disk, false on failure */
bool saveAgent(const UTree &utree, const std::string &filename) {

	if (options.count("binary-io") > 0) {

		std::ofstream ofs(filename.c_str(), std::ios_base::binary);
		if (!ofs) return false;
 
		boost::archive::binary_oarchive oa(ofs);
        oa << utree;

	} else {
		
		std::ofstream ofs(filename.c_str());
		if (!ofs) return false;
    
		boost::archive::text_oarchive oa(ofs);
        oa << utree;
    }

	LINFO_ << "successfully saved agent to: " << filename;
	return true;
}


/* handle user commands (prefixed by colons), returning
   false if input is not a user command. */
static bool dispatchCommand(const std::string &cmd) {

	typedef boost::tokenizer<boost::char_separator<char> > Tok;
	boost::char_separator<char> sep(" ");
	Tok tok(cmd, sep);

	if (cmd.size() == 0 || cmd[0] != ':') return false;

	if (cmd == ":quit") {

		std::exit(0);

	} else if (cmd.substr(0, 5) == ":save") {

		Tok::iterator it = tok.begin();
		std::string filename = "";
		if (++it != tok.end()) filename = *it;
		
		assert(utree.get() != NULL);
		bool rval = saveAgent(*utree, filename);
		if (!rval) {
			LINFO_ << "failed to save agent at " << filename;
		}

	} else {

		LINFO_ << "unrecognised user command, ignoring...";
		return false;
	}

	return true;
}


/* determine the action from a mixture of the U-Tree statistics
   and our exploration policy. */
static action_t getAction(double explore_rate) {

	action_t action;
	if (urng() < explore_rate) {
		double r = urng() * double(utree->nActions());
		action = action_t(r);
		LINFO_ << "Exploring action: " << action;
	} else {
		action = utree->bestAction();
		LINFO_ << "Greedily choosing action: " << action;
	}
	return action;
}


/* agent <-> environment communication loop */
static void mainLoop(void) {

	unsigned int obs_bits = options["observation-bits"].as<unsigned int>();
	unsigned int rew_bits = options["reward-bits"].as<unsigned int>();
	unsigned int act_bits = num_action_bits();
	unsigned int percept_bits = obs_bits + rew_bits;

	double learning_rate  = options["learning-rate"].as<double>();
	double discount       = options["discount"].as<double>();

	double exploration    = options["exploration"].as<double>();
	double exp_decay      = options["explore-decay"].as<double>();

	unsigned int n_actions    = options["agent-actions"].as<unsigned int>();

	unsigned int fringe_depth = options["fringe"].as<unsigned int>();
	unsigned int test_fringe  = options["test-fringe"].as<unsigned int>();

	unsigned int max_back     = options["depth"].as<unsigned int>();
	unsigned int term_age     = options["terminate-age"].as<unsigned int>();

	double total_reward = 0.0;
	
	std::string buf;
	action_t last_action = 0;
	age_t timestep = 1;

	// create U-Tree
	if (options.count("agent-load") > 0) {
		bool use_binary_io = options.count("binary-io") > 0;
		utree.reset(new UTree(options["agent-load"].as<std::string>(), use_binary_io));
	} else {
		utree.reset(new UTree(obs_bits, n_actions, max_back, discount));
	}

	while (std::getline(std::cin, buf)) {

		// process any user directives
		if (dispatchCommand(buf)) continue;

		// grab percept
		boost::dynamic_bitset<> p = boost::dynamic_bitset<>(buf);
		if (p.size() != percept_bits) throw BadPerceptException();

		// split into observation and reward components
		boost::dynamic_bitset<> rew = (p << obs_bits) >> obs_bits;
		boost::dynamic_bitset<> obs = (p >> rew_bits);
		rew.resize(rew_bits); obs.resize(obs_bits);

		LINFO_ << "Received reward of " << rew.to_ulong() << " on timestep " << timestep;
		double avg_reward = total_reward / double(timestep);
		LINFO_ << "Average session reward: " << avg_reward;

		// add new transition instance to the history
		instance_t instance(last_action, obs, rew.to_ulong(), timestep);
		utree->update(instance);
		//assert(utree.integrityTest());
		total_reward += instance.reward;
		
		// perform a single sweep of value iteration
		size_t nodes_updated = utree->sweep();
		LINFO_ << "Value iteration updated " << nodes_updated << " node/s";

		// decide on an action
		LINFO_ << "Exploring with probability: " << exploration;
		
		// keep track of some statistics for this cycle
		logStatistics(*utree);

		// determine whether we need to split the current state representation
		if (timestep % test_fringe == 0) {
			LINFO_ << "Testing fringe of U-Tree for further utile distinctions";
			unsigned int k = utree->trySplits(fringe_depth);
			LINFO_ << k << " splits made";
		}

		LINFO_ << "Currently in State " << utree->state();

		action_t action = getAction(exploration);

		// send response to the environment
		boost::dynamic_bitset<> response(act_bits, action);
		response.resize(act_bits);
		std::cout << response << std::endl;
		
		// move to next iteration
		last_action = action;
		timestep++;
		exploration *= exp_decay;

		if (timestep > term_age) break;
	}

	// display session summary
	std::ostringstream oss;
	oss << "Session summary: " << total_reward 
		<< " reward from " << (timestep-1) << " cycles.";
	LINFO_ << oss.str();
	std::cout << oss.str();
}


/* application entry point */
int main(int argc, char *argv[]) {

	try {

		initLogs();

		// load program configuration
        initOptions(argc, argv, options); 

        // check and apply program configuration
        processOptions(options);

		// enter agent <-> environment communication loop
		mainLoop();
		
	} catch (std::exception &e) {
    
		LERR_ << "error: " << e.what() << std::endl;
        return 1;

    } catch (...) {

		LERR_ << "error: unknown exception" << std::endl;
        return 1;
    }

	LINFO_ << "agent terminated";

	return 0;
}



