#ifndef __UTREE_HPP__
#define __UTREE_HPP__

#include <exception>
#include <vector>
#include <map>
#include <set>

// boost includes
#include <boost/random.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/ptr_container/serialize_ptr_vector.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/dynamic_bitset.hpp> // unofficial


// program configuration options
namespace boost {
	namespace program_options {
		class variables_map;
	}
}
extern boost::program_options::variables_map options;


// describe the reward accumulated by an agent
typedef double reward_t;

// describe the age of an agent
typedef unsigned int age_t;

// observation type
typedef boost::dynamic_bitset<> observation_t;

// describes an agent action
typedef unsigned int action_t;

// random number generator to supply noise
typedef boost::mt19937 randsrc_t;
typedef boost::uniform_01<randsrc_t> randgen_t;

// node identifier type
typedef int unode_id_t;


// describe an instance of the form (a_{t-1}, o_t, r_t, t)
struct instance_t {

	friend class boost::serialization::access;

	instance_t(action_t a, observation_t o, reward_t r, int t) 
		: action(a), obs(o), reward(r), timestep(t) { }

	// serialization routine
	template<class Archive>
	void serialize(Archive &ar, const unsigned int version);

	action_t action;
	observation_t obs;
	reward_t reward;
	unsigned int timestep;

	private:
		// only used by boost serialization
		instance_t(void) { }
};
std::ostream &operator<<(std::ostream &o, const instance_t &instance);

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void instance_t::serialize(Archive &ar, const unsigned int version) {
	
	ar & action;
	ar & obs;	
	ar & reward; 
	ar & timestep; 
}


// describes the accumulated agent experience
typedef std::vector<instance_t> history_t;


// describe a Utile distinction
struct distinction_t {

	friend class boost::serialization::access;
	
	distinction_t(void) :
		back_idx(-1) { }
	distinction_t(int back, int dim)
		: back_idx(back), dimension(dim) { }

	// serialization routine
	template<class Archive>
	void serialize(Archive &ar, const unsigned int version);

	bool operator==(const distinction_t &rhs) {
		return back_idx == rhs.back_idx && dimension == rhs.dimension;
	}

	int back_idx;
	int dimension;
};
std::ostream &operator<<(std::ostream &o, const distinction_t &d);

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void distinction_t::serialize(Archive &ar, const unsigned int version) {
	
	ar & back_idx;
	ar & dimension;
}


// holds statistics about a given action from this leaf
typedef std::map<unode_id_t, int> trans_cnt_t;
struct action_model_t {

	friend class boost::serialization::access;

	action_model_t(void) : count(0), reward(0.0), q_value(0.0) {}

	// serialization routine
	template<class Archive>
	void serialize(Archive &ar, const unsigned int version);

	int count;              // how many times action has been tried
	double reward;          // total accumulated reward
	double q_value;         // estimated q-value
	trans_cnt_t trans_cnt;  // store transition statistics
};
std::ostream &operator<<(std::ostream &o, const action_model_t &am);

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void action_model_t::serialize(Archive &ar, const unsigned int version) {
	
	ar & count;
	ar & reward;
	ar & q_value;
	ar & trans_cnt;
}

typedef std::map<action_t, action_model_t> model_t;


// a node in a U-Tree
class UNode {

	friend class boost::serialization::access;

	friend class UTree;
	friend std::ostream &operator<<(std::ostream &o, const UNode &un);

	public:

		UNode(int id, int type, UNode *parent);

		// compute the expected utility of a leaf
		double utility(void) const;

		// compute the probabililty of transitioning to particular
		// state if a certain action is taken
		double transProb(unode_id_t id, action_t action) const;

		// compute the expected immediate reward by taking some action
		double expectedReward(action_t action) const;

		// adds an instance to the current leaf node
		void insert(const instance_t &instance);

		// get the leaf matching the current state
		UNode *getLeaf(const history_t &history) const;

		// recursively perform a sweep of value iteration on all
		// the leaf nodes contained in the current subtree
		size_t sweep(double discount);

		// determine the best action
		action_t bestAction(void) const;

		// determine the id of a node
		unode_id_t id(void) const;

		// update the transition and reward models
		void updateModel(action_t a, unode_id_t id, double reward);

		// number of children nodes
		size_t nChildren(void) const;

		// number of instances within a node
		size_t nInstances(void) const;

		// determine the parent of a node
		const UNode *parent(void) const;

		// recursively send information about children nodes to stream
		void childrenToStream(std::ostream &o) const;

		// the type of node
		int type(void) const;

		// serialization routine
		template<class Archive>
		void serialize(Archive &ar, const unsigned int version);

	private:

		// only used by Boost Serialization
		UNode(void) { }

		// determine the relevant child of a split node by
		// applying a utile distinction
		size_t applyDistinction(const history_t &history, int idx) const;

		UNode *m_parent;
		unode_id_t m_id;
		int m_type;
		
		model_t m_model;
		distinction_t m_distinction;
		boost::ptr_vector<UNode> m_children;
		std::vector<instance_t> m_instances;
};
std::ostream &operator<<(std::ostream &o, const UNode &un);

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void UNode::serialize(Archive &ar, const unsigned int version) {
	
	ar & m_parent;
	ar & m_id;
	ar & m_type;
	ar & m_model;
	ar & m_distinction;
	ar & m_children;
	ar & m_instances;
}



// implements McCallum's U-Tree
class UTree {

	friend class boost::serialization::access;

	friend std::ostream &operator<<(std::ostream &o, const UTree &utree);

	typedef std::vector<distinction_t> distinction_list_t;
	typedef std::vector<double> efdr_list_t;
	typedef std::map<unode_id_t, UNode *> node_list_t;

	public:

		UTree(unsigned int obs_bits, unsigned int n_actions, 
		  unsigned int max_back, double discount);

		// create U-Tree from a file
		UTree(const std::string &filename, bool use_binary_io);

		// perform a single sweep of value iteration
		// across all the leaf nodes, return the number of
		// nodes updated
		size_t sweep(void);

		// processes a new instance
		void update(const instance_t &instance);

		// return the best action
		action_t bestAction(void) const;

		// get the model statistics for a particular action
		const action_model_t *actionModel(action_t a) const;

		// expand the fringe, returning the number of helpful splits
		unsigned int trySplits(unsigned int fringe);

		// the size of the action space
		unsigned int nActions(void) const;

		// how many observation bits in each percept
		unsigned int nObsBits(void) const;

		// maximum depth back that a utile distinction can be made
		unsigned int maxBackDepth(void) const;

		// access the root U-Node
		const UNode &root(void) const;

		// the id of the current state
		unode_id_t state(void) const;

		// checks the integrity of the u-tree, false on failure
		bool integrityTest(void);

		// serialization routine
		template<class Archive>
		void serialize(Archive &ar, const unsigned int version);

	private:

		// expected future discounted reward associated with an instance
		double expectedFutureReward(const instance_t &inst);

		// the expected future discounted reward of all of a nodes instances
		void getEDFRList(UNode &un, efdr_list_t &efdr_list);
	
		// attempt to find a worthwhile utile distinction from a current node
		distinction_t *getUtileDistinction(UNode &un, distinction_list_t &dl);

		// attempts to split nodes according to a utile distinction test
		unsigned int trySplitRecursive(UNode &un, unsigned int fringe);

		// get the list of ancestor distinctions 
		void ancestorDistinctions(const UNode &un, distinction_list_t &occurred) const;

		// generate all possibile utile distinctions
		void genSplitCandidates(const UNode &un, distinction_list_t &candidates) const;

		// transform a leaf node into a split node
		void split(UNode &un, const distinction_t &d);

		// transform a (recently) split node into a leaf node, inverse of split
		void unsplit(UNode &un);

		// generate a new unique id
		unode_id_t genId(void);

		// determine the current node corresponding
		// the the agent's state
		UNode *getLeaf(void) const;

		// get the leaf node associated with an instance
		UNode *getInstanceLeaf(const instance_t &inst);

		// compute the transition probability for a node from scratch
		double transProbFromInstances(const UNode &un, unode_id_t id, action_t action);

		// compute the instanteous reward for a node from scratch
		double rewardFromInstances(const UNode &un, action_t action);

		// compute the model from the instances
		void modelFromInstances(UNode &un);

		// checks the integrity of the u-tree, false on failure
		bool integrityTestWorker(UNode &n);

		// the instance that was received after the current one
		const instance_t *nextInstance(const instance_t &inst) const;

		UNode m_root;
		unode_id_t m_node_id_c;
		unsigned int m_obs_bits;
		unsigned int m_n_actions;
		unsigned int m_max_back_depth;
		history_t m_history;
		double m_discount;
		node_list_t m_nodes;
};
std::ostream &operator<<(std::ostream &o, const UTree &utree);

/* serialization routine - needs to be accessible from header */
template<class Archive>
inline void UTree::serialize(Archive &ar, const unsigned int version) {
	
	ar & m_root;
	ar & m_node_id_c;
	ar & m_obs_bits;
	ar & m_n_actions;
	ar & m_max_back_depth;
	ar & m_history;
	ar & m_discount;
	ar & m_nodes;
}


#endif // __UTREE_HPP__



